@@ -4,7 +4,7 @@
  *	  Module to handle changes that took place while new table was being
  *	  created
  *
- * Copyright (c) 2016-2023, CYBERTEC PostgreSQL International GmbH
+ * Copyright (c) 2016-2024, CYBERTEC PostgreSQL International GmbH
  *
  *-----------------------------------------------------------------------------------
  */
@@ -19,9 +19,14 @@
 #include "replication/decode.h"
 #include "utils/rel.h"
 
+#if PG_VERSION_NUM < 150000
+extern PGDLLIMPORT int wal_segment_size;
+#endif
+
 static void apply_concurrent_changes(DecodingOutputState *dstate,
 									 Relation relation, ScanKey key,
-									 int nkeys, IndexInsertState *iistate);
+									 int nkeys, IndexInsertState *iistate,
+									 struct timeval *must_complete);
 static bool processing_time_elapsed(struct timeval *utmost);
 
 static void plugin_startup(LogicalDecodingContext *ctx,
@@ -60,10 +65,20 @@ process_concurrent_changes(LogicalDecodingContext *ctx,
 	bool		done;
 
 	dstate = (DecodingOutputState *) ctx->output_writer_private;
+
+	/*
+	 * If some changes could not be applied due to time constraint, make sure
+	 * the tuplestore is empty before we insert new tuples into it.
+	 */
+	if (dstate->nchanges > 0)
+		apply_concurrent_changes(dstate, rel_dst, ident_key,
+								 ident_key_nentries, iistate, NULL);
+	Assert(dstate->nchanges == 0);
+
 	done = false;
 	while (!done)
 	{
-		CHECK_FOR_INTERRUPTS();
+		exit_if_requested();
 
 		done = decode_concurrent_changes(ctx, end_of_wal, must_complete);
 
@@ -83,7 +98,11 @@ process_concurrent_changes(LogicalDecodingContext *ctx,
 		 * non-trivial.
 		 */
 		apply_concurrent_changes(dstate, rel_dst, ident_key,
-								 ident_key_nentries, iistate);
+								 ident_key_nentries, iistate, must_complete);
+
+		if (processing_time_elapsed(must_complete))
+			/* Like above. */
+			return false;
 	}
 
 	return true;
@@ -102,6 +121,11 @@ decode_concurrent_changes(LogicalDecodingContext *ctx,
 {
 	DecodingOutputState *dstate;
 	ResourceOwner resowner_old;
+#if PG_VERSION_NUM < 130000
+	/* Workaround for XLogBeginRead() in setup_decoding(). */
+	static	bool	first_time = true;
+	XLogRecPtr	startptr;
+#endif
 
 	/*
 	 * Invalidate the "present" cache before moving to "(recent) history".
@@ -129,9 +153,19 @@ decode_concurrent_changes(LogicalDecodingContext *ctx,
 			char	   *errm = NULL;
 			XLogRecPtr	end_lsn;
 
+#if PG_VERSION_NUM < 130000
+			if (first_time)
+			{
+				startptr = MyReplicationSlot->data.restart_lsn;
+				first_time = false;
+			}
+			else
+				startptr = InvalidXLogRecPtr;
+#endif
+
 			record = XLogReadRecord(ctx->reader,
 #if PG_VERSION_NUM < 130000
-									InvalidXLogRecPtr,
+									startptr,
 #endif
 									&errm);
 			if (errm)
@@ -157,7 +191,7 @@ decode_concurrent_changes(LogicalDecodingContext *ctx,
 				squeeze_current_segment = segno_new;
 			}
 
-			CHECK_FOR_INTERRUPTS();
+			exit_if_requested();
 		}
 		InvalidateSystemCaches();
 		CurrentResourceOwner = resowner_old;
@@ -185,12 +219,11 @@ decode_concurrent_changes(LogicalDecodingContext *ctx,
  */
 static void
 apply_concurrent_changes(DecodingOutputState *dstate, Relation relation,
-						 ScanKey key, int nkeys, IndexInsertState *iistate)
+						 ScanKey key, int nkeys, IndexInsertState *iistate,
+						 struct timeval *must_complete)
 {
 	TupleTableSlot *slot;
-#if PG_VERSION_NUM >= 120000
 	TupleTableSlot *ind_slot;
-#endif
 	Form_pg_index ident_form;
 	int2vector *ident_indkey;
 	HeapTuple	tup_old = NULL;
@@ -204,17 +237,11 @@ apply_concurrent_changes(DecodingOutputState *dstate, Relation relation,
 	ident_indkey = &ident_form->indkey;
 
 	/* TupleTableSlot is needed to pass the tuple to ExecInsertIndexTuples(). */
-#if PG_VERSION_NUM >= 120000
 	slot = MakeSingleTupleTableSlot(dstate->tupdesc, &TTSOpsHeapTuple);
-#else
-	slot = MakeSingleTupleTableSlot(dstate->tupdesc);
-#endif
 	iistate->econtext->ecxt_scantuple = slot;
 
-#if PG_VERSION_NUM >= 120000
 	/* A slot to fetch tuples from identity index. */
 	ind_slot = table_slot_create(relation, NULL);
-#endif
 
 	/*
 	 * In case functions in the index need the active snapshot and caller
@@ -225,9 +252,7 @@ apply_concurrent_changes(DecodingOutputState *dstate, Relation relation,
 	while (tuplestore_gettupleslot(dstate->tstore, true, false,
 								   dstate->tsslot))
 	{
-#if PG_VERSION_NUM >= 120000
 		bool		shouldFree;
-#endif
 		HeapTuple	tup_change,
 					tup,
 					tup_exist;
@@ -236,12 +261,11 @@ apply_concurrent_changes(DecodingOutputState *dstate, Relation relation,
 		bool		isnull[1];
 		Datum		values[1];
 
+		Assert(dstate->nchanges > 0);
+		dstate->nchanges--;
+
 		/* Get the change from the single-column tuple. */
-#if PG_VERSION_NUM >= 120000
 		tup_change = ExecFetchSlotHeapTuple(dstate->tsslot, false, &shouldFree);
-#else
-		tup_change = ExecFetchSlotTuple(dstate->tsslot);
-#endif
 		heap_deform_tuple(tup_change, dstate->tupdesc_change, values, isnull);
 		Assert(!isnull[0]);
 
@@ -283,19 +307,12 @@ apply_concurrent_changes(DecodingOutputState *dstate, Relation relation,
 			heap_insert(relation, tup, GetCurrentCommandId(true), 0, bistate);
 
 			/* Update indexes. */
-#if PG_VERSION_NUM >= 120000
 			ExecStoreHeapTuple(tup, slot, false);
-#else
-			ExecStoreTuple(tup, slot, InvalidBuffer, false);
-#endif
 			recheck = ExecInsertIndexTuples(
 #if PG_VERSION_NUM >= 140000
 											iistate->rri,
 #endif
 											slot,
-#if PG_VERSION_NUM < 120000
-											&(tup->t_self),
-#endif
 											iistate->estate,
 #if PG_VERSION_NUM >= 140000
 											false,	/* update */
@@ -319,9 +336,9 @@ apply_concurrent_changes(DecodingOutputState *dstate, Relation relation,
 			pfree(tup);
 
 			/* Update the progress information. */
-			SpinLockAcquire(&MyWorkerProgress->mutex);
-			MyWorkerProgress->ins += 1;
-			SpinLockRelease(&MyWorkerProgress->mutex);
+			SpinLockAcquire(&MyWorkerSlot->mutex);
+			MyWorkerSlot->progress.ins += 1;
+			SpinLockRelease(&MyWorkerSlot->mutex);
 		}
 		else if (change->kind == PG_SQUEEZE_CHANGE_UPDATE_NEW ||
 				 change->kind == PG_SQUEEZE_CHANGE_DELETE)
@@ -370,7 +387,6 @@ apply_concurrent_changes(DecodingOutputState *dstate, Relation relation,
 												  &isnull);
 				Assert(!isnull);
 			}
-#if PG_VERSION_NUM >= 120000
 			if (index_getnext_slot(scan, ForwardScanDirection, ind_slot))
 			{
 				bool		shouldFreeInd;
@@ -382,9 +398,7 @@ apply_concurrent_changes(DecodingOutputState *dstate, Relation relation,
 			}
 			else
 				tup_exist = NULL;
-#else
-			tup_exist = index_getnext(scan, ForwardScanDirection);
-#endif
+
 			if (tup_exist == NULL)
 				elog(ERROR, "Failed to find target tuple");
 			ItemPointerCopy(&tup_exist->t_self, &ctid);
@@ -402,15 +416,25 @@ apply_concurrent_changes(DecodingOutputState *dstate, Relation relation,
 								   &update_indexes
 #endif
 					);
+				/*
+				 * In PG < 16, change of any indexed attribute makes HOT
+				 * impossible, Therefore HOT update implies that no index
+				 * needs to be updated.
+				 *
+				 * In PG >= 16, if only attributes of "summarizing indexes"
+				 * change, HOT update is still possible. Therefore HOT update
+				 * might still require some indexes (in particular, the
+				 * summarizing ones) to be updated.
+				 */
+#if PG_VERSION_NUM >= 160000
+				if (update_indexes != TU_None)
+#else
 				if (!HeapTupleIsHeapOnly(tup))
+#endif
 				{
 					List	   *recheck;
 
-#if PG_VERSION_NUM >= 120000
 					ExecStoreHeapTuple(tup, slot, false);
-#else
-					ExecStoreTuple(tup, slot, InvalidBuffer, false);
-#endif
 
 					/*
 					 * XXX Consider passing update=true, however it requires
@@ -422,9 +446,6 @@ apply_concurrent_changes(DecodingOutputState *dstate, Relation relation,
 													iistate->rri,
 #endif
 													slot,
-#if PG_VERSION_NUM < 120000
-													&(tup->t_self),
-#endif
 													iistate->estate,
 #if PG_VERSION_NUM >= 140000
 													false,	/* update */
@@ -442,18 +463,18 @@ apply_concurrent_changes(DecodingOutputState *dstate, Relation relation,
 				}
 
 				/* Update the progress information. */
-				SpinLockAcquire(&MyWorkerProgress->mutex);
-				MyWorkerProgress->upd += 1;
-				SpinLockRelease(&MyWorkerProgress->mutex);
+				SpinLockAcquire(&MyWorkerSlot->mutex);
+				MyWorkerSlot->progress.upd += 1;
+				SpinLockRelease(&MyWorkerSlot->mutex);
 			}
 			else
 			{
 				simple_heap_delete(relation, &ctid);
 
 				/* Update the progress information. */
-				SpinLockAcquire(&MyWorkerProgress->mutex);
-				MyWorkerProgress->del += 1;
-				SpinLockRelease(&MyWorkerProgress->mutex);
+				SpinLockAcquire(&MyWorkerSlot->mutex);
+				MyWorkerSlot->progress.del += 1;
+				SpinLockRelease(&MyWorkerSlot->mutex);
 			}
 
 			if (tup_old != NULL)
@@ -474,15 +495,25 @@ apply_concurrent_changes(DecodingOutputState *dstate, Relation relation,
 			UpdateActiveSnapshotCommandId();
 		}
 
-#if PG_VERSION_NUM >= 120000
 		/* TTSOpsMinimalTuple has .get_heap_tuple==NULL. */
 		Assert(shouldFree);
 		pfree(tup_change);
-#endif
+
+		/*
+		 * If there is a limit on the time of completion, check it
+		 * now. However, make sure the loop does not break if tup_old was set
+		 * in the previous iteration. In such a case we could not resume the
+		 * processing in the next call.
+		 */
+		if (must_complete && tup_old == NULL &&
+			processing_time_elapsed(must_complete))
+			/* The next call will process the remaining changes. */
+			break;
 	}
 
-	tuplestore_clear(dstate->tstore);
-	dstate->nchanges = 0;
+	/* If we could not apply all the changes, the next call will do. */
+	if (dstate->nchanges == 0)
+		tuplestore_clear(dstate->tstore);
 
 	PopActiveSnapshot();
 
@@ -490,9 +521,7 @@ apply_concurrent_changes(DecodingOutputState *dstate, Relation relation,
 	if (bistate != NULL)
 		FreeBulkInsertState(bistate);
 	ExecDropSingleTupleTableSlot(slot);
-#if PG_VERSION_NUM >= 120000
 	ExecDropSingleTupleTableSlot(ind_slot);
-#endif
 }
 
 static bool
@@ -645,7 +674,11 @@ plugin_change(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 				HeapTuple	newtuple;
 
 				newtuple = change->data.tp.newtuple != NULL ?
+#if PG_VERSION_NUM >= 170000
+					change->data.tp.newtuple : NULL;
+#else
 					&change->data.tp.newtuple->tuple : NULL;
+#endif
 
 				/*
 				 * Identity checks in the main function should have made this
@@ -663,9 +696,17 @@ plugin_change(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 							newtuple;
 
 				oldtuple = change->data.tp.oldtuple != NULL ?
+#if PG_VERSION_NUM >= 170000
+					change->data.tp.oldtuple : NULL;
+#else
 					&change->data.tp.oldtuple->tuple : NULL;
+#endif
 				newtuple = change->data.tp.newtuple != NULL ?
+#if PG_VERSION_NUM >= 170000
+					change->data.tp.newtuple : NULL;
+#else
 					&change->data.tp.newtuple->tuple : NULL;
+#endif
 
 				if (newtuple == NULL)
 					elog(ERROR, "Incomplete update info.");
@@ -681,7 +722,11 @@ plugin_change(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 				HeapTuple	oldtuple;
 
 				oldtuple = change->data.tp.oldtuple ?
+#if PG_VERSION_NUM >= 170000
+					change->data.tp.oldtuple : NULL;
+#else
 					&change->data.tp.oldtuple->tuple : NULL;
+#endif
 
 				if (oldtuple == NULL)
 					elog(ERROR, "Incomplete delete info.");
